using System;
using System.Threading;
using SDL2;
using Serilog;
using Serilog.Core;
using SlatedGameToolkit.Framework.Exceptions;
using SlatedGameToolkit.Framework.StateSystem;

namespace SlatedGameToolkit.Framework {
    /// <summary>
    /// The main engine that will host the game loop.
    /// </summary>
    public static class GameEngine {
        public static readonly Logger logger = new LoggerConfiguration().
        WriteTo.File("SlatedGameToolKit.Framework.Log", rollingInterval: RollingInterval.Day, fileSizeLimitBytes: 1048576, rollOnFileSizeLimit: true).
        CreateLogger();
        private static readonly object ignitionLock = new object();
        private static readonly object deltaUpdateLock = new object();
        private static Thread thread;
        private static volatile bool exit = false, stopped = true;
        private static volatile bool deltaChanged;
        private static TimeSpan updateDeltaTime = TimeSpan.FromSeconds(1d/200d), frameDeltaTime = TimeSpan.FromSeconds(1d/60d);

        /// <summary>
        /// The amount of updates per second.
        /// Is floored to milleseconds.
        /// Maximum updates per second is 200.
        /// </summary>
        /// <value>The updates per second.</value>
        public static double UpdatesPerSecond {
            get {
                return  1d / updateDeltaTime.TotalSeconds;
            }
            set {
                lock (deltaUpdateLock) {
                    if (1 / value < 0.005) {
                        updateDeltaTime = TimeSpan.FromSeconds(1d/200d);
                    } else {
                        updateDeltaTime = TimeSpan.FromSeconds(1d/value);
                    }
                }
                deltaChanged = true;
            }
        }

        /// <summary>
        /// The target frames per second. Will not go above this number, but may dip below this number.
        /// This value is floored to millesconds units.
        /// Setting this to 0 renders as many frames as the system can handle.
        /// Default is 60.
        /// </summary>
        /// <value>The target frames per second.</value>
        public static double targetFPS {
            get {
                return 1 / frameDeltaTime.TotalSeconds;
            }
            set {
                lock (deltaUpdateLock) {
                    if (value == 0) {
                        frameDeltaTime = TimeSpan.FromSeconds(0);
                    } else {
                        frameDeltaTime = TimeSpan.FromSeconds(1d/value);
                    }
                }
                deltaChanged = true;
            }
        }

        public static bool Running {
            get {
                return !stopped;
            }
        }

        private static void Loop(Object o) {
            if (!(o is Manager)) throw new InternalFrameworkException(String.Format("Expected manager object for asynchronous loop. Got {0}", o));
            Manager manager = (Manager) o;
            manager.initialize();
            DateTime currentTime = DateTime.Now;
            TimeSpan timePassedFromLastUpdate;
            TimeSpan timePassedFromLastRender;
            TimeSpan updateDeltaTime;
            TimeSpan frameDeltaTime;
            deltaChanged = true;
            stopped = false;
            logger.Information("Game engine initiated.");
            while (!exit) {
                if (deltaChanged) {
                    lock (deltaUpdateLock) {
                        updateDeltaTime = GameEngine.updateDeltaTime;
                        frameDeltaTime = GameEngine.frameDeltaTime;
                        logger.Information(String.Format("Deltas were set. Update Delta: {0}, Render Delta: {1}", updateDeltaTime.TotalSeconds, frameDeltaTime.TotalSeconds));
                    }
                    deltaChanged = false;
                }

                DateTime frameStart = DateTime.Now;
                TimeSpan difference = frameStart - currentTime;
                currentTime = frameStart;
                
                timePassedFromLastUpdate += difference;
                while (timePassedFromLastUpdate > updateDeltaTime) {
                    manager.update(updateDeltaTime.TotalMilliseconds);
                    timePassedFromLastUpdate -= updateDeltaTime;
                }

                timePassedFromLastRender += difference;
                if (timePassedFromLastRender > frameDeltaTime) {
                    manager.render(timePassedFromLastUpdate.TotalMilliseconds / updateDeltaTime.TotalMilliseconds);
                    timePassedFromLastRender = TimeSpan.Zero;
                }
            }
            manager.removeAllStates();
            stopped = true;
            SDL.SDL_Quit();
            logger.Information("Game engine has stopped.");
        }

        /// <summary>
        /// Requests to stop the game engine.
        /// Game engine will finish the frame its currently in.
        /// This needs to be called to perform proper clean up.
        /// </summary>
        /// <returns>False if the stop request has already been made.</returns>
        public static bool Stop() {
            lock (ignitionLock) {
                if (exit) return false;
                exit = true;
                return true;
            }
        }

        /// <summary>
        /// Requests to start the game engine. 
        /// </summary>
        /// <returns>True iff the engine is not already running.</returns>
        public static bool Ignite(Manager manager) {
            if (manager == null) throw new ArgumentNullException("manager");
            SDL.SDL_version SDLVersion;
            SDL.SDL_version SDLBuiltVersion;
            SDL.SDL_GetVersion(out SDLVersion);
            SDL.SDL_VERSION(out SDLBuiltVersion);
            logger.Information(String.Format("Attempting to initiate game engine with SDL version: {0}.{1}.{2}", SDLVersion.major, SDLVersion.minor, SDLVersion.patch));
            if (!SDL.SDL_VERSION_ATLEAST(SDLBuiltVersion.major, SDLBuiltVersion.minor, SDLBuiltVersion.patch)) {
                logger.Warning(String.Format("Engine was designed with SDL version {0}.{1}.{2}, currently running on {3}.{4}.{5}", SDLBuiltVersion.major, SDLBuiltVersion.minor, SDLBuiltVersion.patch, SDLVersion.major, SDLVersion.minor, SDLVersion.patch));
                if (SDLVersion.major < 2) {
                    logger.Error("This engine was designed to work with SDL2. The version you're currently running is severely outdated.");
                    throw new FrameworkUsageException("Outdated SDL binaries.");
                }
            }
            lock (ignitionLock) {
                if (!stopped) {
                    logger.Warning("Engine is already running.");
                    return false;
                }
                exit = false;
                if (SDL.SDL_Init(SDL.SDL_INIT_VIDEO) != 0) {
                    throw new FrameworkSDLException();
                }
                if (SDL.SDL_Init(SDL.SDL_INIT_AUDIO) != 0) {
                    throw new FrameworkSDLException();
                }

                if (SDL.SDL_GL_SetAttribute(SDL.SDL_GLattr.SDL_GL_CONTEXT_MAJOR_VERSION, 3) < 0) throw new FrameworkSDLException();
                if (SDL.SDL_GL_SetAttribute(SDL.SDL_GLattr.SDL_GL_CONTEXT_MINOR_VERSION, 1) < 0) throw new FrameworkSDLException();
                if (SDL.SDL_GL_SetAttribute(SDL.SDL_GLattr.SDL_GL_CONTEXT_PROFILE_MASK, SDL.SDL_GLprofile.SDL_GL_CONTEXT_PROFILE_CORE) < 0) throw new FrameworkSDLException();

                thread = new Thread(Loop);
                thread.Start(manager);
                return true;
            }
        }
    }
}